/*
 * database.c (c) 2002 by Benjamin Schweizer <gopher at h07 dot org>
 *                        http://www.redsheep.de/
 *
 **********************************************************************/


#include <stdio.h>
#include <pcap.h>				// pcap sniffer lib
#include <string.h>

#include "globals.h"
#include "database.h"


/* locals *
 *********/
 
struct shost db[DB_SIZE];


/* implementation *
 ******************/

void tochar(u_char *c, const char *s) {
/*
 * tochar(out, in) - converts hex into byte
 */
	if (s[0] > 96) {				// a-f
		*c = (s[0] - 87) << 4;
	} else if (s[0] > 64) {				// A-F
		*c = (s[0] - 55) << 4;
	}  else {					// 0-9
		*c = (s[0] - 48) << 4;
	}
	if (s[1] > 96) {				// a-f
		*c |= (s[1] - 87);
	} else if (s[1] > 64) {				// A-F
		*c |= (s[1] - 55);
	} else {					// 0-9
		*c |= (s[1] - 48);
	}
}

int db_save (const char *name) {
/*
 * db_save(filename) - stores data
 */
	FILE *fp;
	int i;

	if ((fp = fopen(name, "w")) == NULL) {
		return (-1);
	}
	for (i = 0; i < db_last; i++) {			// errorhandling
		fprintf (fp, "%02x%02x%02x%02x%02x%02x:%s\n",
			db[i].mac[0], db[i].mac[1], db[i].mac[2],
			db[i].mac[3], db[i].mac[4], db[i].mac[5],
			db[i].desc);
	}
	fclose(fp);

	return 0;
}

int db_load (const char *name) {
/*
 * db_load(filename) - loads data
 */
	FILE *fp;
	char buff[1024];
	char *sp;
	char c = 0;

	db_last = 0;
	
	if ((fp = fopen(name, "r")) == NULL) {
		return (-1);
	}
	while (fgets(buff, sizeof(buff), fp) != (char *) NULL) {
		if ((sp = strchr(buff, '\n')) != (char *) NULL) {
			*sp = '\0'; }

    		tochar(&c, buff + 0);			// decode hex
		db[db_last].mac[0] = c;
    		tochar(&c, buff + 2);
		db[db_last].mac[1] = c;
    		tochar(&c, buff + 4);
		db[db_last].mac[2] = c;
    		tochar(&c, buff + 6);
		db[db_last].mac[3] = c;
    		tochar(&c, buff + 8);
		db[db_last].mac[4] = c;
    		tochar(&c, buff + 10);
		db[db_last].mac[5] = c;

		db[db_last].seen = 0;			// mark unseen
		if (sizeof(buff) > 13) {
			memcpy(db[db_last].desc, buff + 13, sizeof(buff) - 13);
		} else {
			db[db_last].desc[0] = '\0';
		}

		if (db_last <= DB_SIZE) db_last++;
	}

	fclose(fp);
	return 0;
}

int db_check (struct shost *entry, u_char *old_ip) {
/*
 * db_check(entry) - compare/update entry with db
 */
	int i;
	for (i = 0; i < db_last; i++) {
		if (memcmp (db[i].mac, entry->mac, 6) == 0) {
			strcpy (entry->desc, db[i].desc);
			if (db[i].seen == 0) {		// firsttime
				memcpy (db[i].ipv4, entry->ipv4, 4);
				memcpy (db[i].ipv6, entry->ipv6, 16);
			        db[i].seen = 1;		// mark as seen
				return DBMSG_UNSEEN;
			} else {			// already known
			        if (memcmp (entry->ipv4, "\0\0\0\0", 4) != 0) {	// check and update if ipv4 address changed
					if (memcmp (db[i].ipv4, entry->ipv4, 4) != 0) {
						memcpy (old_ip, db[i].ipv4, 4);
						memcpy (db[i].ipv4, entry->ipv4, 4);
						return DBMSG_IPV4;
					}
				}
    				if (memcmp (entry->ipv6, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 16) != 0) {	// check and update if ipv6 address changed
					if (memcmp (db[i].ipv6, entry->ipv6, 16) != 0) {
						memcpy (old_ip, db[i].ipv6, 16);
						memcpy (db[i].ipv6, entry->ipv6, 16);
						return DBMSG_IPV6;
					}
				}
				return DBMSG_ALIVE;	// already seen
			}
		}
	}
	return DBMSG_NEW;				// new host
}

int db_add (const struct shost *entry) {
/*
 * db_add(entry) - add entry to db
 */
	memcpy (db[db_last].mac, entry->mac, 6);
	memcpy (db[db_last].ipv4, entry->ipv4, 4);
	memcpy (db[db_last].ipv6, entry->ipv6, 16);

	strcpy(db[db_last].desc, entry->desc);
	db[db_last].seen = 1;					// mark as seen
	if (db_last < DB_SIZE) {
		db_last++;
	} else {
		// printf("warning: db full!\n");
		return -1;
	};
	return 0;					// new host
}


/* end of file */
