/*
 * echolot.c (c) 2002 by Benjamin Schweizer <gopher at h07 dot org>
 *                       http://www.redsheep.de/
 *
 **********************************************************************/


#include <stdio.h>
#include <pcap.h>				// pcap sniffer lib
#include <string.h>
#include <sys/signal.h>				// sighandler
#include <getopt.h>				// cmdline
#include <netdb.h>				// nslookups
#include <sys/socket.h>
#include <syslog.h>

#include "globals.h"
#include "echolot.h"
#include "database.h"
#include "ui.h"


/* locals *
 *********/
 
static pcap_t *pd;
int snaplen = DEFAULT_SNAPLEN;
char *db_name, *db_desc, *device, *ext_script;

static struct proto protos[] = {			// only ethernet
        { ethernet,	DLT_EN10MB },
        { NULL,		0 },
};


/* implementation *
 ******************/

void finish (void) {
/*
 * finish - signal handler
 */
	char messagebuffer[1024];
 	pcap_close(pd);
	
	log (MSG_NOTICE, "\n\nfinished sniffing.\n");
	if (o_add) {
		sprintf (messagebuffer, "writing %s...", db_name);
		log (MSG_NOTICE, messagebuffer);
		if (db_save(db_name) == -1) {
			log (MSG_WARNING, "can't write database.\n\n");
		} else {
			sprintf (messagebuffer, "%d entries done.\n\n", db_last);
			log (MSG_IMPORTANT, messagebuffer);
		}
	} else {
		log (MSG_NOTICE, "not saving any entries, use --add parameter\n\n");
	}
	if (o_daemon) {
		syslog (LOG_NOTICE, "program terminated.");
		remove (PIDFILE);
	}
	closelog;
	exit(0);
}

void set_signal(int signo, void (*handler)(void)) {
/*
 * set_signal - setup signal handler
 */
        struct sigaction sa;
        memset(&sa, 0, sizeof(sa));
        sa.sa_handler = (void (*)(int))handler;
        sa.sa_flags = SA_RESTART;
        sigaction(signo, &sa, NULL);
}

int other_instance() {
/*
 * other_instance - check if echolot is already running
 */
        FILE *fp;

	if ((fp = fopen(PIDFILE, "r")) != NULL) {
	        fclose(fp);
		return 1;
	}
	return 0;
}

void daemonize () {
/*
 * daemonize - check/create pid file and daemonize
 */
        FILE *fp;
	int pid;

	if (other_instance() && !o_force) {
	        printf("echolot is already running, pidfile %s exists.\n", PIDFILE);
		exit(1);
	}
	if ((fp = fopen(PIDFILE, "w")) == NULL) {
		printf("cannot open %s.\n", PIDFILE);
		exit(1);
	}

	pid = fork();
	if (pid < 0) {
		printf("cannot fork.\n", PIDFILE);
	        fclose(fp);
		exit(1);
	}
	if (pid != 0) {	// we are parent, we terminate now.
		exit(1);
	}
	fprintf (fp, "%d\n", getpid());
        fclose(fp);
}

struct arp_packet decode_packet (const u_char *dp) {
/*
 * decode_packet
 */
	struct arp_packet darp;
	memcpy (darp.dest, dp + 0, 6);		// ethernet frame
	memcpy (darp.src, dp + 6, 6);
	memcpy (darp.type, dp + 12, 2);

	memcpy (darp.hrd, dp + 14, 2);		// packet
	memcpy (darp.pro, dp + 16, 2);
	memcpy (darp.hlen, dp + 18, 1);
	memcpy (darp.plen, dp + 19, 1);
	memcpy (darp.op, dp + 20, 2);

//printf ("hlen = %d, plen = %d\n", *darp.hlen, *darp.plen);
	// on startup, there could be crazy values...
	if (*darp.hlen <= 6 && *darp.plen <= 16) {
		memcpy (darp.sha, dp + 22, *darp.hlen);	// more packet, beware that address size may be different
		memcpy (darp.spa, dp + 22 + *darp.hlen, *darp.plen);
		memcpy (darp.tha, dp + 22 + *darp.hlen + *darp.plen, *darp.hlen);
		memcpy (darp.tpa, dp + 22 + *darp.hlen + *darp.plen + *darp.hlen, *darp.plen);
	} else {
		darp.hlen[0] = 0;
		darp.plen[0] = 0;
	}
	return darp;
}

void ethernet(u_char *user, const struct pcap_pkthdr *header, const u_char *raw_packet) {
/*
 * ethernet - handle new packets
 */
	struct arp_packet packet;
	u_char old_ip[16];	// old ip address if db_check returns ip change
	char x_src[18];		// ethernet hw address in human readable format
	char x_sha[18];		// ethernet hw address in human readable format, may
	char x_ip[40];		// ipv4/6 address in human readable format
	char messagebuffer[1024];

	struct shost host;
	struct hostent *addr;
	int ret;
	
	FILE *external;
	char retstr[DB_DLEN];
	char callstr[255];

	packet = decode_packet(raw_packet);
	
	if (*packet.hlen == 6) {		// we are on ethernet
		sprintf(x_src, "%02x:%02x:%02x:%02x:%02x:%02x", packet.src[0], packet.src[1], packet.src[2], packet.src[3], packet.src[4], packet.src[5]);
		sprintf(x_sha, "%02x:%02x:%02x:%02x:%02x:%02x", packet.sha[0], packet.sha[1], packet.sha[2], packet.sha[3], packet.sha[4], packet.sha[5]);
		if (*packet.plen = 4) {	// plen=4 looks like ipv4
			sprintf(x_ip, "%d.%d.%d.%d", packet.spa[0], packet.spa[1], packet.spa[2], packet.spa[3]);
			memcpy (host.ipv4, packet.spa, 4);
			memcpy (host.ipv6, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 16);
		} else {
			if (*packet.plen = 16) {	// plen=16 looks like ipv6
				sprintf (x_ip, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
				    packet.spa[0], packet.spa[1], packet.spa[2], packet.spa[3],
				    packet.spa[4], packet.spa[5], packet.spa[6], packet.spa[7],
				    packet.spa[8], packet.spa[9], packet.spa[10], packet.spa[11],
				    packet.spa[12], packet.spa[13], packet.spa[14], packet.spa[15]);
				memcpy (host.ipv4, "\0\0\0\0", 4);
				memcpy (host.ipv6, packet.spa, 16);
			} else {
				sprintf (x_ip, "non-ip");
				memcpy (host.ipv4, "\0\0\0\0", 4);
				memcpy (host.ipv6, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 16);
				printf ("fetched non ipv4/6 packet with plen = %s\n", packet.plen);
			}
		}
		memcpy (host.mac, packet.sha, *packet.hlen);
		strncpy (host.desc, db_desc, DB_DLEN);

		// process packet
		if ((ret = db_check(&host, old_ip)) > 0) {		// we already know this host
			if (ret == DBMSG_UNSEEN) {			// 1st time
				sprintf(messagebuffer, "%s is already known (%s, %s)\n", x_src, host.desc, x_ip);
				log (MSG_IMPORTANT, messagebuffer);
			}
			if (o_cont && ret == DBMSG_ALIVE) {		// we already announced this host,  --continuously
				sprintf(messagebuffer, "%s is still alive (%s, %s)\n", x_src, host.desc, x_ip);
				log (MSG_IMPORTANT, messagebuffer);
			}
			if (ret == DBMSG_IPV4) {			// changed ipv4 address
				sprintf(messagebuffer, "%s changed ip (%s, %d.%d.%d.%d -> %s)\n", x_src, host.desc, old_ip[0], old_ip[1], old_ip[2], old_ip[3], x_ip);
				log (MSG_IMPORTANT, messagebuffer);
			}
			if (ret == DBMSG_IPV6) {			// changed ipv6 address
				sprintf(messagebuffer, "%s changed ip (%s, %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x -> %s)\n",
				    x_src, host.desc,
				    old_ip[0], old_ip[1], old_ip[2], old_ip[3],
				    old_ip[4], old_ip[5], old_ip[6], old_ip[7],
				    old_ip[8], old_ip[9], old_ip[10], old_ip[11],
				    old_ip[12], old_ip[13], old_ip[14], old_ip[15], x_ip);
				log (MSG_IMPORTANT, messagebuffer);
			}
		} else {	// this is a new host
			if ((*packet.plen = 4) && o_script) {	// resolve name (ipv4), this may take a little
					/*	// internal name server code
					 * addr = gethostbyaddr(packet.spa, 4, AF_INET);
					 * if (addr != NULL) {		// ns failed?
					 * 	strncpy (host.desc, addr->h_name, DB_DLEN);
					 * } else {
					 * 	strncpy (host.desc, db_desc, DB_DLEN);
					 * }
					 */
				sprintf (callstr, "%s %s %s", ext_script, x_ip, db_desc);
				external = popen (callstr, "r");
				if (fgets (retstr, DB_DLEN, external) != (char *) NULL) {
					strncpy (host.desc, retstr, DB_DLEN);
				} else {			// maybe script failed
					strncpy (host.desc, db_desc, DB_DLEN);
				}
				pclose (external);
			}
			if (db_add(&host) == -1) { log (MSG_WARNING, "warning: database is full!\n"); };		// check with db
			sprintf(messagebuffer, "%s is unknown as yet (%s, %s)\n", x_src, host.desc, x_ip);
			log (MSG_IMPORTANT, messagebuffer);
		}
		if (memcmp (packet.src, packet.sha, *packet.hlen) != 0) {	// compare hw address and given address
			sprintf(messagebuffer, "%s tried to spoof %s, %s.\n", x_src, x_sha, x_ip);
			log (MSG_IMPORTANT, messagebuffer);
		}
	} // drop non-arp packets
}

static pcap_handler lookup_callback(int type)		// only ethernet
{
        struct proto *p;

        for (p = protos; p->f; ++p)
                if (type == p->type)
                        return p->f;

        printf("unknown data link type %d", type);
	return NULL;
        /* NOTREACHED */
}

void version() {
	// we are printing version at startup
	printf (
		"copyright (c) 2002 by Benjamin Schweizer <gopher at h07 dot org>\n"
		"                      http://www.redsheep.de/\n"
		"\n"
		"Redistribution and use in source and binary forms, with or without\n"
		"modification, are permitted provided that the following conditions\n"
		"are met:\n"
		"\n"  
		"  1. Redistributions of source code must retain the above copyright\n"
		"     notice, this list of conditions and the following disclaimer.\n"
		"  2. Redistributions in binary form must reproduce the above copyright\n"
		"     notice, this list of conditions and the following disclaimer in\n"
		"     the documentation and/or other materials provided with the\n"
		"     distribution.\n"
		"  3. The names of the authors may not be used to endorse or promote\n"
		"     products derived from this software without specific prior\n"
		"     written permission.\n"
		"\n"
		"THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR\n"
		"IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED\n"
		"WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.\n\n\n"
	);
	exit(0);
}

void usage() {
	printf( "\n"
		"echolot [-a][-c][-D][-f] [-i interface] [-d database] [-I idendity] [-s default.pl] [-h|-v]\n"
		"\n"
		"	-a|--add		add new hosts to database\n"
		"	-c|--continuously	announce continuously\n"
		"	-D|--daemonize		start as a daemon and write to the syslog\n"
		"	-f|--force		force startup if pid file already exists\n"
		"\n"
		"	-i|--interface		use given interface, default is \"any\"\n"
		"	-d|--database		use given database, default is \"/var/db/echolot.mac\"\n"
		"\n"
		"	-I|--idendity		set default description for new hosts\n"
		"	-s|--script		use external resolver script, i.e. \"echolot -s scripts/default.pl\"\n"
 		"\n"
		"	-h|--help		print this short help\n"
       		"	-v|--version		print release and copyright information\n\n"
	);
	exit(0);
}


/* entry *
 ********/
 
int main (int argc, char **argv) {
	register int i;
	bpf_u_int32 localnet, netmask;
	pcap_handler callback;
	struct bpf_program fcode;
	char errbuf[PCAP_ERRBUF_SIZE];
	char messagebuffer[1024];
	int lop;

	struct option longopts[] = {
		{"add", 0, 0, 'a'},
		{"continuously", 0, 0, 'c'},
		{"daemonize", 0, 0, 'D'},
		{"force", 0, 0, 'f'},
		{"interface", 1, 0, 'i'},
		{"database", 1, 0, 'd'},
		{"idendity", 1, 0, 'I'},
		{"script", 1, 0, 's'},
		{"help", 0, 0, 'h'},
		{"version", 0, 0, 'v'},
		{NULL, 0, 0, 0}
	};

	o_add = 0; o_cont = 0; o_script = 0; o_daemon = 0; o_force = 0;	// defaults
	device = NULL;
	db_name = DATABASE;
	db_desc = "default";
	ext_script = "scripts/default.pl";

	openlog ("echolot", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_DAEMON);	// init syslog for daemon mode
	printf ("\n%s\n\n", RELEASE);

	while ((i = getopt_long(argc, argv, "acDfi:d:I:s:hv", longopts, &lop)) != EOF)
		switch (i) {
		case 'a':
			o_add = 1;
			break;
		case 'c':
			o_cont = 1;
			break;
		case 'D':
			o_daemon = 1;
			break;
		case 'f':
			o_force = 1;
			break;
		case 'i':
			device = optarg;
			break;
		case 'd':
			db_name = optarg;
			break;
		case 'I':
			db_desc = optarg;
			break;
		case 's':
			ext_script = optarg;
	 		o_script = 1;
			break;
		case 'v':
			version();
			break;
		case '?':
		case 'h':
		default:
			usage();
		}

	if (o_daemon && other_instance() && !o_force) {		// no two daemons
	        printf("echolot is already running, pidfile %s exists.\n", PIDFILE);
		exit(1);
	}
	printf ("loading %s... ", db_name);
	if (db_load(db_name) == -1) { printf ("can't read database.\n"); };
	printf ("%d entries done.\n", db_last);

	if (device == NULL) {				// get default
		device = pcap_lookupdev(errbuf);
		if (device == NULL) { printf("%s\n", errbuf); exit(1); }
	}
	pd = pcap_open_live(device, snaplen, 1, 10, errbuf);
	if (pd == NULL) { printf("%s", errbuf); exit(1); }
	i = pcap_snapshot(pd);
	if (pcap_lookupnet(device, &localnet, &netmask, errbuf) < 0) { 
		localnet = 0;				// network down?
		netmask = 0;
		printf("%s", errbuf);
		exit(1);
	}

	netmask = 0;					// we take everything we can get;)
        if (pcap_compile(pd, &fcode, "arp", 1, netmask) < 0)	// arp is enough
		{ printf("%s", pcap_geterr(pd)); exit(1); }
	if (pcap_setfilter(pd, &fcode) < 0)
		{ printf("%s", pcap_geterr(pd)); exit(1); }

	printf("collecting data on %s.\n\n", device);

	set_signal(SIGINT, finish);			// signal handle (^C)
	set_signal(SIGTERM, finish);			// signal handle (TERM)
	callback = lookup_callback(pcap_datalink(pd));	// net handle

	if (o_daemon) {
		printf ("daemonizing...\n\n");
		daemonize();
		syslog (LOG_NOTICE, "%s started.", RELEASE);
	}
        if (pcap_loop(pd, -1, callback, 0) < 0) {	// run handle
                sprintf(messagebuffer, "\nsniffing failed: %s\n", pcap_geterr(pd));	// maybe interface is down?
		log (MSG_WARNING, messagebuffer);
		if (o_add) {
			sprintf (messagebuffer, "writing %s...", db_name);
			log (MSG_NOTICE, messagebuffer);
			if (db_save(db_name) == -1) {
				log (MSG_WARNING, "can't write database.\n\n");
			} else {
				sprintf (messagebuffer, "%d entries done.\n\n", db_last);
				log (MSG_NOTICE, messagebuffer);
			}
		}
		if (o_daemon) {
			remove (PIDFILE);
		}
                exit(1);
        }

	/*
	 *  we will hang in the pcap_loop and exit over sigint-handler.
	 */
	log (MSG_WARNING, "FATAL: program terminated abnormaly. pcap_loop exited abnormaly.");
	if (o_add) {
		sprintf (messagebuffer, "writing %s...", db_name);
		log (MSG_NOTICE, messagebuffer);
		if (db_save(db_name) == -1) {
			log (MSG_WARNING, "can't write database.\n\n");
		} else {
			sprintf (messagebuffer, "%d entries done.\n\n", db_last);
			log (MSG_NOTICE, messagebuffer);
		}
	}
	if (o_daemon) {
		remove (PIDFILE);
	}
	
	pcap_close(pd);
	closelog;
	return -1;
}

/* end of file */
